//
// utils.cc: simple utility classes (linked list and hashing functions).
//
// ------------------------------------------------
// Mumit Khan <khan@xraylith.wisc.edu>
// Center for X-ray Lithography
// University of Wisconsin-Madison
// 3731 Schneider Dr., Stoughton, WI, 53589
// ------------------------------------------------
//
// Copyright (c) 1991-1996 Mumit Khan
//
//

/*****************************************************************************
 *
 *  INCLUDES
 *
 ****************************************************************************/


#ifdef __GNUG__
# include <cassert>
# include <cctype>
# include <cstdio>
# include <cstdlib>
# include <cstring>
# include <iostream.h>
#else
# include <assert>
# include <ctype>
# include <stdio>
# include <stdlib>
# include <string>
# include <iostream>
#endif

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef __BORLANDC__
# define R_OK    4               /* Test for read permission.  */
# define W_OK    2               /* Test for write permission.  */
# define X_OK    1               /* Test for execute permission.  */
# define F_OK    0               /* Test for existence.  */
# include <io.h>
#endif

#include <fcntl.h>
#include <sys/stat.h>

#include "global.h"
#include "utils.h"

/************************************************************************/

// NOTE: USER RESPONSIBLE FOR MEMORY MANAGEMENT OF item
void Stq::push (void const* item) {
    Stqitem* lp;
    if (free_ == nil) 
	lp = new Stqitem();
    else { 
	lp = free_;  
	lp->item = lp->next = nil;  
	free_ = nil; 
    }
    lp->item = (void*)item;	// UNCONST
    if (head_ == nil) 
	tail_ = lp;
    else 
	lp->next = head_;
    head_ = lp;
    nitems_++;
}

void* Stq::get (int ith) const {
    assume (ith > 0 && ith <= nitems_);
    void* ret = nil;
    Stqitem* lp = head_;
    int j = 0;
    while (lp != nil) { 
	j++; 
	if (ith == j) {
	    ret =  lp->item;  
	    break;
	}
	else {
	    lp = lp->next;  
	}
    }
    return ret;
}


// NOTE: USER RESPONSIBLE FOR MEMORY MANAGEMENT OF item
void* Stq::pop () {
    void* item;
    Stqitem* lp;
    if (head_ == nil) 
	return nil;
    item = head_->item;
    lp = head_;
    head_ = head_->next;
    if (head_ == nil) 
	tail_ = nil;
    if (free_ == nil) 
	free_ = lp;  
    else 
	delete lp;
    nitems_--;
    return item;
}

// NOTE: USER RESPONSIBLE FOR MEMORY MANAGEMENT OF item
void Stq::enq (void const* item) {
    Stqitem* lp;
    if (free_ == nil) 
	lp = new Stqitem();
    else { 
	lp = free_;  lp->item = lp->next = nil;  free_ = nil; 
    }
    lp->item = (void*)item;	// UNCONST
    if (tail_ == nil) 
	head_ = lp;
    else 
	tail_->next = lp;
    tail_ = lp;
    nitems_++;
}

void* Stq::cycle() const { 
    Stq* thislist  = (Stq*)this;
    void* item = thislist->pop(); thislist->enq(item); 
    return item; 
}

void Stq::insert (int ith, void const* item) {
    if(ith == 1) {
	push(item);
    }
    else if (ith == nitems_ + 1) {
	enq(item);
    }
    else {
	Stqitem* lp;
	if (free_ == nil) 
	    lp = new Stqitem();
	else { 
	    lp = free_;  lp->item = lp->next = nil;  free_ = nil; 
	}
	lp->item = (void*)item;	// UNCONST

	// get the item no. ith-1
	Stqitem* prev = head_;
	for (unsigned i = 1; prev && i < ith-1; prev = prev->next, i++)
	    ;
	assume(prev != nil);
	lp->next = prev->next;
	prev->next = lp;
	nitems_++;
    }
}

void* Stq::remove (int ith) {
    assume (ith > 0 && ith <= nitems_);
    if (ith == 1) {
	return pop();
    }

    // get the item no. ith-1
    Stqitem* prev = head_;
    for (unsigned i = 1; prev && i < ith-1; prev = prev->next, i++)
	;
    assume(prev != nil);
    Stqitem* lp = prev->next;
    void* item = lp->item;
    prev->next = lp->next;
    if (prev->next == nil)
	tail_ = prev;

    if (free_ == nil)
	free_ = lp;
    else
	delete lp;
    nitems_--;

    return item;
}

Stq* Stq::clone() const {
    Stq* buddy = new Stq();
    Stqitem* hp = head_;
    while (hp != nil) { 
	buddy->enq(hp->item);
	hp = hp->next;
    }
    return buddy;
}

// NOTE: ONLY Stqitems ARE DELETED...NOT items ADDED BY USERS
void Stq::clear() {
    for (int i=1;i<=nitems_;i++) 
	pop();
    if (free_ != nil) 
	delete free_;
}

/**************************************************************************/

Hashtable::Hashtable(int tablesize) {
    tablesize_ = tablesize;
    table_ = new Hashlink*[tablesize_];
    for (int i = 0; i < tablesize_; i++) 
	table_[i] = nil;
    nitems_ = 0;
}

Hashtable::~Hashtable() {  
    clear();  
}

void Hashtable::clear() {
    for (int i = 0; i < tablesize_; i++) {
	Hashlink* cp = table_[i];
	while (cp != nil) { 
	    Hashlink* lp = cp; 
	    cp = cp->next;  
	    // for debugging .... CHECK
	    lp->key = nil; lp->item = nil; lp->next = nil;
	    delete lp;
	}
	table_[i] = nil;
    }
    nitems_ = 0; 
}

int Hashtable::hash(char const* key) const {
    unsigned g;  unsigned h = 0;
    for (char const* s=key; *s != '\0'; s++) {
	h = (h << 4) + (toupper(*s));
	if ((g = h&0xf0000000)) { 
	    h = h ^ (g >> 24);  h = h ^ g; 
	}
    }
    return (h % tablesize_);
}

bool Hashtable::compare (char const* key1, char const* key2) const {
    bool isequal = false;
    char const* k1 = key1;  char const* k2 = key2;
    for (; toupper(*k1) == toupper(*k2); k1++, k2++) {
	if (*k1 == '\0' && *k2 == '\0') {
	    isequal = true;
	    break;
	}
    }
    return isequal;
}

void Hashtable::insert (char const* key, void const* item) {
    int result = hash (key);
    Hashlink* cp = new Hashlink(key, item);
    cp->next = table_[result];
    table_[result] = cp;
    nitems_++;
}

void* Hashtable::find (char const* key) const {
    void const* item = nil;
    int result = hash (key);
    for (Hashlink* cp = table_[result]; cp; cp = cp->next) {
	if (compare(key, cp->key)) {
	    item = cp->item;
	    break;
	}
    }
    return (void*)item;			// UNCONST
}

void* Hashtable::remove (char const* key) {
    int result = hash(key);
    Hashlink* cp = table_[result];  Hashlink* lp = nil;
    while (cp != nil) {
	if (compare(key, cp->key)) {
	    if (lp != nil) 
		lp->next = cp->next;
	    else 
		table_[result] = cp->next;
	    void const* item = cp->item;
	    // for debugging .... CHECK
	    cp->key = nil; cp->item = nil; cp->next = nil;
	    delete cp;
	    nitems_--;
	    return (void*)item;		// UNCONST
	}
	lp = cp;  cp = cp->next;
    }
    return nil;
}
void* Hashtable::get(int ith) const {
    assume (ith > 0);  assume (ith <= nitems_);
    int count = 0; 
    for (int i = 0; i <= tablesize_; i++) {
	if (table_[i] != nil) {
	    Hashlink* cp = table_[i];
	    while (cp != nil) { 
		count++;  
		if (count == ith) 
		    return (void*)cp->item; // UNCONST
		cp = cp->next;
	    } 
	}
    }
    assume (0);
    return nil;
}

Hashtable* Hashtable::clone (COPIER_FP clonerfunc) const {
    Hashtable* buddy = new Hashtable(tablesize_);
    for (int i=0; i<=tablesize_; i++) {
	if (table_[i] != nil) {
	    Hashlink* cp = table_[i];
	    while (cp != nil) { 
		buddy->insert(
		    cp->key, (clonerfunc) ? (*clonerfunc)(cp->item) : cp->item
		);
		cp = cp->next;
	    }
	}
    }
    return buddy;
}

/*******************************************************************************
*
*  OTHER UTILITIES
*
*******************************************************************************/

// REPLACE THE STRING replacethis WITH withthin IN inthisline
// PUT THE RESULTS IN results
// RETURN POSIITON IN results OF 1st CHAR WHERE THE SUBSTITUTION TOOK PLACE
// OR 0 IF THERE WAS NO SUBSTITUTION
// IF NO SUBSTITUTION TAKES PLACE results CONTAINS A COPY OF inthisline
// MATCH MUST BE EXACT..."Hi" DOES NOT MATCH "hi"
int strsub (
    const char* replacethis, const char* withthis, const char* inthisline, 
    char* results
) {
    int withlen,replacelen,subpos,resultspos,ret,stop;
    char const* k1; 
    char const* k2;

   strcpy (results, inthisline);
   replacelen = strlen (replacethis);  if (replacelen == 0) return 0;
   withlen = strlen (withthis);
   stop = strlen (inthisline) - replacelen;
   subpos = resultspos = ret = 0;
   while (1) {
      k1 = replacethis;  k2 = &inthisline[subpos];
      for (;*k1==*k2;k1++,k2++) if (*k1 == '\0') break;
      if (*k1 == '\0') {
         // MATCH...COPY INTO results AND CONTINUE
         // NOTE: k2 IS ONE PAST THE MATCH CHARS
         strncpy (&results[resultspos],withthis,withlen);
         ret = resultspos;
         resultspos += withlen-1;  subpos += replacelen-1;
         strcpy (&results[resultspos+1],&inthisline[subpos+1]);
      }
      subpos++;  resultspos++;  if (subpos > stop) break;
   }
   return ret;
}

//
// expandpath: expand a PATH by looking at embedded env variables.
// doesn't handle ~ (tilde) expansion yet, but will someday.
//

int expandpath(const char* orig_path, char path[]) {
    if (GLB_debug) {
	cerr << "expandpath: Expanding " << orig_path << " ... ";
    }
    char env_var[1024];
    char tmppath[1024];
    strcpy(path, orig_path);
    strcpy(tmppath, orig_path);
    while(1) {
	char* start = strchr(path, '$');
	char* end = start;
	for(; end && *end && *end != '/' && *end != ' ' && *end != '\t'; ++end)
	    ;
	if (end == start)
	    break;
	strncpy(env_var, start, end-start);
	env_var[end-start] = '\0';
	char* variable = &env_var[1];
	const char* env_value;
	if (!(env_value = getenv(variable))) {
	    cerr << endl << "ERROR: environment variable `" << variable
		<< "' not found." << endl;
	    return 1;
	}
	strsub(env_var, env_value, tmppath, path);
	strcpy(tmppath, path);
    }
    if (GLB_debug) {
	cerr << path << endl;
    }
    return 0;
}

//
// find_include_file: look for the include file in the set of standard
// search directories supplied.
//

int find_include_file(
    const char* orig_path, const Stq* search_path, char path[]
) {
    if (GLB_debug) {
	cerr << "find_include_file: searching for " << orig_path << " ... ";
    }
    bool found = false;
    unsigned ndirs = search_path->size();
    while(ndirs--) {
	if (found)
	    continue;		// busy loop to finish cycle. work done.
	char buf[1024];
	const char* directory = (const char*)search_path->cycle();
	sprintf(buf, "%s/%s", directory, orig_path);
	if (fileexists(buf)) {
	    found = true;
	    strcpy(path, buf);
	}
    }
    return (found) ? 0 : 1;
}

//
// search_file: scans the filename and figures out if it should be 
// searched for in a set of standard places or not. If so, scan and
// return the full pathname where found. Uses find_include_file.
//
// return code:
//    0: AOK
//    1: couldn't find file anywhere
//    2: embedded env variable not found
//

int search_file(
    const char* orig_path, const Stq* search_path, char path[]
) {
    char pathname[1024];
    int error_code = 0;
    bool do_search = false;
    if (orig_path[0] == '<' && orig_path[strlen(orig_path)-1] == '>') {
	//
	// this is standard include directive, so set a flag
	// we scan later, except when the filename has a 
	// leading '/', in which case we don't scan.
	//
	const int namelen = strlen(orig_path) - 2;
	pathname[namelen] = '\0';
	strncpy(pathname, &orig_path[1], namelen);
	do_search = true;
    } else {
	strcpy(pathname, orig_path);
    }
    //
    // now expand embedded env vars.
    //
    if (expandpath(pathname, path)) {
	error_code = 2;
    } else {
	if (do_search && path[0] != '/') {
	    strcpy(pathname, path);
	    if (find_include_file(
		pathname, search_path, path
	    )) {
		error_code = 1;
	    }
	} 
    }
    return error_code;
}

void getunusedpath (char* path, char* prefix, bool create) {
// COPY INTO path A FILE NAME THAT IS NOT CURRENTLY BEING USED
// IF prefix IS NOT nil USE IT AS A PREFIX
// IF create == 1 CREATE THE FILE (EMPTY)
char buf[1024];
char buf2[128];
int  suffix = 0;
   // KEEP CONSTRUCTING FILES TILL YOU GET A UNIQUE ONE
   while (1) {
      strcpy  (buf,prefix);
      sprintf (buf2,"%d",suffix);  suffix++;
      strcat (buf,buf2);
      if (access(buf,F_OK)==-1) break;
   }
   strcpy (path,buf);
   if (create) { int fp = creat (path,0777); close (fp); }
}

char* getprogramname (char* argv[]) {
// RETURN POINTER TO PROGRAM NAME
   
   return argv[0];
}
int getflag (char flag, int argc, char* argv[]) {
// RETURN 1 IF FLAG FOUND; OTHERWISE RETURN 0
   for (int i=0;i<argc;i++) 
      if (argv[i][0] == '-') 
         if (argv[i][1] == flag) return 1;
   return 0;
}
int getflagval (char flag, int argc, char* argv[], char* buf) {
// RETURN VALUE OF FLAG flag INTO buf
// RETURN 1 IF FLAG FOUND; OTHERWISE RETURN 0
   strcpy (buf,"");
   for (int i=0;i<argc;i++) 
      if (argv[i][0] == '-') if (argv[i][1] == flag) if (i+1 < argc)
         { strcpy (buf,argv[i+1]); return 1; }
   return 0;
}
void promptforval (const char* prompt, char* buf) {
// PROMPT USER FOR INPUT (prompt) AND PUT RESULT IN buf
// strlen(buf) == 0 MEANS THERE WAS NO INPUT
   strcpy (buf,"");
   printf ("%s",prompt);
   int pos = 0;
   while (1) { 
      int ch = getchar();
      if (ch == EOF) break;
      if (ch == 13) break;
      if (ch == 10) break;
      buf[pos++] = ch;
    }
   buf[pos] = '\0';
}
bool fileexists (const char* path) {
// RETURN true IF FILE EXISTS, OTHERWISE RETURNS false
struct stat sb;
int type;
   if (path == nil) return 0;  if (strlen(path) == 0) return 0;
   if (stat(path,&sb) < 0) return 0;
   // MAKE SURE ITS A REGULAR FILE (NOT A DIRECTORY)
   type = ((sb.st_mode & S_IFMT) == S_IFREG);  if (type == 0) return 0;
   return 1;
}

char* strcpy_with_alloc (char const* str) {
    char* dup = nil;
    if (str) {
	char* copy = new char[::strlen(str) + 1];
	::strcpy (copy, str);
	dup = copy;
    }
    return dup;
}

void assume_fail (const char* x, const char* file, int line)
{
    fprintf (stderr, "The assertion '%s' failed: FILE=%s LINE=%d.\n",x,file,line);
    assert (0);
}
